# coding: utf-8
from typing import TYPE_CHECKING, Dict, Optional

from .endpoint import APIEndpoint
from .models import User

if TYPE_CHECKING:
    from .client import NuxeoClient


class API(APIEndpoint):
    """Endpoint for users."""

    __slots__ = ()

    def __init__(self, client, endpoint="user", headers=None):
        # type: (NuxeoClient, str, Optional[Dict[str, str]]) -> None
        super().__init__(client, endpoint=endpoint, cls=User, headers=headers)

    def get(self, user_id=None, ssl_verify=True):
        # type: (Optional[str], Optional[bool]) -> User
        """
        Get the detail of a user.

        :param user_id: the id of the user
        :return: the user
        """
        return super().get(path=user_id, ssl_verify=ssl_verify)

    def post(self, user, ssl_verify=True):
        # type: (User, bool) -> User
        """
        Create a user.

        :param user: the user to create
        :return: the created user
        """
        return super().post(user, ssl_verify=ssl_verify)

    create = post  # Alias for clarity

    def put(self, user, ssl_verify=True):
        # type: (User, bool) -> User
        """
        Update a user.

        :param user: the user to update
        :return: the updated user
        """
        return super().put(user, ssl_verify=ssl_verify)

    def delete(self, user_id, ssl_verify=True):
        # type: (str, bool) -> None
        """
        Delete a user.

        :param user_id: the id of the user to delete
        """
        super().delete(user_id, ssl_verify=ssl_verify)

    def current_user(self, ssl_verify=True):
        # type: (bool) -> User
        """
        Get the current user details and validate the connection to the server at the same time.

        :return User: user's details
        """
        details = self.client.request(
            "GET", "site/api/v1/me", ssl_verify=ssl_verify
        ).json()
        return User(**details)



# coding: utf-8
from base64 import b64encode
from typing import Optional
from ..exceptions import NuxeoError
from ..utils import get_digest_hash


def make_portal_sso_token(timestamp, random, secret, username, digest_algorithm="md5"):
    # type: (int, str, str, str, Optional[str]) -> str
    """Generate a token for SSO with Portals."""
    digester = get_digest_hash(digest_algorithm)
    if not digester:
        err = f"Cannot compute token because of unknown digest algorithm: {digest_algorithm!r}"
        raise NuxeoError(err)

    clear_token = ":".join([str(timestamp), random, secret, username])
    digester.update(clear_token.encode("utf-8"))
    hashed_token = digester.digest()
    return b64encode(hashed_token).decode("utf-8")



# coding: utf-8
import base64
from requests import Request

from .base import AuthBase


class BasicAuth(AuthBase):
    """ Attaches a simple Basic Authentication to the given Request object. """

    __slots__ = ("username", "password", "_token_header")

    AUTHORIZATION = "Authorization".encode("utf-8")

    def __init__(self, username, password):
        # type: (str, str) -> None
        self.username = username
        self.password = password
        self.set_token(password)

    def set_token(self, token):
        # type: (str) -> None
        """Apply the given *token*."""
        self.password = token
        payload = f"{self.username}:{self.password}".encode("utf-8")
        self._token_header = f"Basic {base64.b64encode(payload).decode('utf-8')}"

    def __eq__(self, other):
        # type: (object) -> bool
        return all(
            [
                self.username == getattr(other, "username", None),
                self.password == getattr(other, "password", None),
            ]
        )

    def __ne__(self, other):
        # type: (object) -> bool
        return not self == other

    def __call__(self, r):
        # type: (Request) -> Request
        r.headers[self.AUTHORIZATION] = self._token_header
        return r



# coding: utf-8
from typing import Any, Dict, List, Optional

from requests.exceptions import RetryError


class NuxeoError(Exception):
    """Mother class for all exceptions."""


class BadQuery(NuxeoError):
    """
    Exception thrown either by an operation failure:
        - when the command is not valid
        - on unexpected parameter
        - on missing required parameter
        - when a parameter has not the required type
    Or by any requests to the server with invalid data.
    """


class CorruptedFile(NuxeoError):
    """Exception thrown when digests of a downloaded blob are different."""

    def __init__(self, filename, server_digest, local_digest):
        # type: (str, str, str) -> None
        self.filename = filename
        self.server_digest = server_digest
        self.local_digest = local_digest

    def __repr__(self):
        # type: () -> str
        return (
            f"CorruptedFile {self.filename!r}: server digest is {self.server_digest!r},"
            f" local digest is {self.local_digest!r}"
        )

    def __str__(self):
        # type: () -> str
        return repr(self)


class HTTPError(RetryError, NuxeoError):
    """Exception thrown when the server returns an error."""

    _valid_properties = {"status": -1, "message": None, "stacktrace": None}

    def __init__(self, **kwargs):
        # type: (Any) -> None
        for key, default in HTTPError._valid_properties.items():
            if key == "status":
                # Use the subclass value, if defined
                value = getattr(self, key, kwargs.get(key, default))
            else:
                value = kwargs.get(key, default)

            setattr(self, key, value)

    def __repr__(self):
        # type: () -> str
        return "%s(%d), error: %r, server trace: %r" % (
            type(self).__name__,
            self.status,
            self.message,
            self.stacktrace,
        )

    def __str__(self):
        # type: () -> str
        return repr(self)

    @classmethod
    def parse(cls, json):
        # type: (Dict[str, Any]) -> HTTPError
        """Parse a JSON object into a model instance."""
        model = cls()

        for key, val in json.items():
            if key in cls._valid_properties:
                setattr(model, key, val)
        return model


class Conflict(HTTPError):
    """Exception thrown when the HTTPError code is 409."""

    status = 409


class Forbidden(HTTPError):
    """Exception thrown when the HTTPError code is 403."""

    status = 403


class InvalidBatch(NuxeoError):
    """Exception thrown when accessing inexistant or deleted batches."""


class InvalidUploadHandler(NuxeoError):
    """Exception thrown when trying to upload a blob using an invalid handler."""

    def __init__(self, handler, handlers):
        # type: (str, List[str]) -> None
        self.handler = handler
        self.handlers = tuple(handlers)

    def __repr__(self):
        # type: () -> str
        return f"{type(self).__name__}: the upload handler {self.handler!r} is not one of {self.handlers}."

    def __str__(self):
        # type: () -> str
        return repr(self)


class OAuth2Error(HTTPError):
    """Exception thrown when an OAuth2 error happens."""

    status = 400

    def __init__(self, error):
        # type: (str) -> None
        self.message = error
        self.stacktrace = None


class OngoingRequestError(Conflict):
    """Exception thrown when doing an idempotent call that is already being processed."""

    def __init__(self, request_uid):
        # type: (str) -> None
        self.request_uid = request_uid

    def __repr__(self):
        # type: () -> str
        return (
            f"{type(self).__name__}: a request with the idempotency key"
            f" {self.request_uid!r} is already being processed."
        )

    def __str__(self):
        # type: () -> str
        return repr(self)


class Unauthorized(HTTPError):
    """Exception thrown when the HTTPError code is 401."""

    status = 401


class UnavailableConvertor(NuxeoError):
    """
    Exception thrown when a converter is registered but not
    available right now (e.g. not installed on the server).

    :param options: options passed to the conversion request
    """

    def __init__(self, options):
        # type: (Dict[str, Any]) -> None
        self.options = options
        self.message = str(self)

    def __repr__(self):
        # type: () -> str
        return f"UnavailableConvertor: conversion with options {self.options!r} is not available"

    def __str__(self):
        # type: () -> str
        return repr(self)


class NotRegisteredConvertor(NuxeoError):
    """
    Exception thrown when a converter is not registered.

    :param options: options passed to the conversion request
    """

    def __init__(self, options):
        # type: (Dict[str, Any]) -> None
        self.options = options
        self.message = str(self)

    def __repr__(self):
        # type: () -> str
        return f"ConvertorNotRegistered: conversion with options {self.options!r} can not be done"

    def __str__(self):
        # type: () -> str
        return repr(self)


class UnavailableBogusConvertor(NuxeoError):
    """
    Exception thrown when a converter is registered but not
    available right now (e.g. not installed on the server)
    or when a converter is not registered
    and an Internal Server Error is thrown instead of actual error message.

    :param error_message: actual error message
    :param converter_name: name of the converter if available else empty string
    """

    def __init__(self, error_message, converter_name):
        # type: (str, str) -> None
        self.error_message = error_message
        self.converter_name = converter_name
        self.message = str(self)

    def __repr__(self):
        # type: () -> str
        msg = (
            "Internal Server Error or Converter "
            + self.converter_name
            + " is not registered or "
            + "UnavailableConvertor: conversion with options or Unsupported Operation"
        )
        return msg

    def __str__(self):
        # type: () -> str
        return repr(self)


class UploadError(NuxeoError):
    """
    Exception thrown when an upload fails even after retries.
    """

    def __init__(self, name, chunk=None, info=None):
        # type: (str, Optional[int], Optional[str]) -> None
        self.name = name
        self.chunk = chunk
        self.info = info

    def __repr__(self):
        # type: () -> str
        err = f"UploadError: unable to upload file {self.name!r}"
        if self.chunk:
            err += f" (failed at chunk {self.chunk})"
        if self.info:
            err += f" (source: {self.info})"
        return err

    def __str__(self):
        # type: () -> str
        return repr(self)



# coding: utf-8
from typing import TYPE_CHECKING, Dict, Optional, Any, List

from .endpoint import APIEndpoint
from .models import Comment

if TYPE_CHECKING:
    from .client import NuxeoClient


class API(APIEndpoint):
    """Endpoint for comments."""

    __slots__ = ()

    def __init__(self, client, endpoint="id", headers=None):
        # type: (NuxeoClient, str, Optional[Dict[str, str]]) -> None
        super().__init__(client, endpoint=endpoint, cls=Comment, headers=headers)

    def get(self, docuid, uid=None, ssl_verify=True, params=None):
        # type: (str, str, bool, Any) -> Comment
        """
        Get the detail of a comment.

        :param uid: the ID of the comment
        :return: the comment
        """
        path = f"{docuid}/@comment"
        if uid:
            path += f"/{uid}"

        # Adding "&fetch-comment=repliesSummary" to the URL to retrieve replies number as well
        kwargs = {"fetch-comment": "repliesSummary"}
        if isinstance(params, dict):
            kwargs.update(params)

        return super().get(ssl_verify=ssl_verify, path=path, params=kwargs)

    def post(self, docuid, text, ssl_verify=True):
        # type: (str, str, bool) -> Comment
        """
        Create a comment.

        :param comment: the comment to create
        :return: the created comment
        """
        kwargs = {"entity-type": "comment", "parentId": docuid, "text": text}

        return super().post(
            ssl_verify=ssl_verify, path=docuid, adapter="comment", resource=kwargs
        )

    create = post  # Alias for clarity

    def put(self, comment, ssl_verify=True):
        # type: (Comment, bool) -> None
        """
        Update a comment.

        :param resource: the entry to update
        :return: the entry updated
        """
        path = f"{comment.parentId}/@comment/{comment.uid}"
        return super().put(resource=comment, path=path, ssl_verify=ssl_verify)

    def delete(self, uid, ssl_verify=True):
        # type: (str, bool) -> None
        """
        Delete a comment.

        :param uid: the ID of the comment to delete
        """
        super().delete(uid, ssl_verify=ssl_verify)

    def replies(self, uid, ssl_verify=True, params=None):
        # type: (str, bool, Any) -> List[Comment]
        """
        Get the replies of the comment.

        Any additionnal arguments will be passed to the *params* parent's call.

        :param uid: the ID of the comment
        :return: the list of replies
        """
        # Adding "&fetch-comment=repliesSummary" to the URL to retrieve replies number as well
        kwargs = {"fetch-comment": "repliesSummary"}
        if isinstance(params, dict):
            kwargs.update(params)

        return super().get(
            path=uid, adapter="comment", ssl_verify=ssl_verify, params=kwargs
        )



# coding: utf-8
import atexit
import json
import logging
from typing import Any, Dict, Optional, Tuple, Type, Union
from warnings import warn

import requests
from requests.adapters import HTTPAdapter
from urllib3 import __version__ as urllib3_version
from urllib3.util.retry import Retry

from . import (
    __version__,
    comments,
    directories,
    documents,
    groups,
    operations,
    tasks,
    uploads,
    users,
    workflows,
)
from .auth.base import AuthBase
from .auth import BasicAuth, TokenAuth
from .constants import (
    CHUNK_SIZE,
    DEFAULT_API_PATH,
    DEFAULT_APP_NAME,
    DEFAULT_URL,
    IDEMPOTENCY_KEY,
    MAX_RETRY,
    RETRY_BACKOFF_FACTOR,
    RETRY_METHODS,
    RETRY_STATUS_CODES,
    TIMEOUT_CONNECT,
    TIMEOUT_READ,
)
from .exceptions import (
    BadQuery,
    Conflict,
    Forbidden,
    HTTPError,
    OngoingRequestError,
    Unauthorized,
)
from .tcp import TCPKeepAliveHTTPSAdapter
from .utils import json_helper, log_response

AuthType = Optional[Union[Tuple[str, str], AuthBase]]
logger = logging.getLogger(__name__)

if urllib3_version < "1.26.0":
    DEFAULT_RETRY = Retry(
        total=MAX_RETRY,
        backoff_factor=RETRY_BACKOFF_FACTOR,
        method_whitelist=RETRY_METHODS,
        status_forcelist=RETRY_STATUS_CODES,
        raise_on_status=False,
    )
else:
    DEFAULT_RETRY = Retry(
        total=MAX_RETRY,
        backoff_factor=RETRY_BACKOFF_FACTOR,
        allowed_methods=RETRY_METHODS,
        status_forcelist=RETRY_STATUS_CODES,
        raise_on_status=False,
    )

# Custom exception to raise based on the HTTP status code
# (default will be HTTPError)
HTTP_ERROR = {
    requests.codes.conflict: Conflict,
    requests.codes.forbidden: Forbidden,
    requests.codes.unauthorized: Unauthorized,
}


class NuxeoClient(object):
    """
    The HTTP client used by Nuxeo.

    :param auth: An authentication object passed to Requests
    :param host: The url of the Nuxeo Platform
    :param api_path: The API path appended to the host url
    :param chunk_size: The size of the chunks for blob download
    :param kwargs: kwargs passed to :func:`NuxeoClient.request`
    """

    def __init__(
        self,
        auth=None,  # type: AuthType
        host=DEFAULT_URL,  # type: str
        api_path=DEFAULT_API_PATH,  # type: str
        chunk_size=CHUNK_SIZE,  # type: int
        **kwargs,  # type: Any
    ):
        # type: (...) -> None
        self.auth = BasicAuth(*auth) if isinstance(auth, tuple) else auth
        self.host = host
        self.api_path = api_path
        self.chunk_size = chunk_size

        version = kwargs.pop("version", "")
        app_name = kwargs.pop("app_name", DEFAULT_APP_NAME)
        self.headers = {
            "X-Application-Name": app_name,
            "X-Client-Version": version,
            "User-Agent": f"{app_name}/{version}",
            "Accept": "application/json, */*",
        }

        self.schemas = kwargs.get("schemas", "*")
        self.repository = kwargs.pop("repository", "default")
        self._session = requests.sessions.Session()
        self._session.hooks["response"] = [log_response]
        cookies = kwargs.pop("cookies", None)
        if cookies:
            self._session.cookies = cookies
        self._session.stream = True
        self.client_kwargs = kwargs

        self.ssl_verify_needed = True
        if "verify" in kwargs:
            self.ssl_verify_needed = kwargs["verify"]

        atexit.register(self.on_exit)

        # Cache for the server information
        self._server_info = None

        # Ensure the host is well formatted
        if not self.host.endswith("/"):
            self.host += "/"

        # The retry adapter
        self.retries = kwargs.pop("retries", None) or DEFAULT_RETRY

        # Install the retries mecanism
        self.enable_retry()

    def __repr__(self):
        # type: () -> str
        return f"{type(self).__name__}<host={self.host!r}, version={self.server_version!r}>"

    def __str__(self):
        # type: () -> str
        return repr(self)

    def on_exit(self):
        # type: () -> None
        self._session.close()

    def enable_retry(self):
        # type: () -> None
        """Set a max retry for all connection errors with an adaptative backoff."""
        self._session.mount(
            "https://", TCPKeepAliveHTTPSAdapter(max_retries=self.retries)
        )
        self._session.mount("http://", HTTPAdapter(max_retries=self.retries))

    def disable_retry(self):
        # type: () -> None
        """
        Restore default mount points to disable the eventual retry
        adapters set with .enable_retry().
        """
        self._session.close()
        self._session.mount("https://", TCPKeepAliveHTTPSAdapter())
        self._session.mount("http://", HTTPAdapter())

    def query(
        self,
        query,  # type: str
        params=None,  # type: Dict[str, Any]
    ):
        """
        Query the server with the specified NXQL query.
        Additional qery parameters can be set via the `params` argument:

            >>> nuxeo.client.query('NXSQL query', params={'properties': '*'})

        You can find what parameters to tweak under the `Repository.Query`
        operation details.
        """

        data = {"query": query}
        if params:
            data.update(params)

        url = f"{self.api_path}/search/lang/NXQL/execute"
        return self.request("GET", url, params=data).json()

    def set(self, repository=None, schemas=None):
        # type: (Optional[str], Optional[str]) -> NuxeoClient
        """
        Set the repository and/or the schemas for the requests.

        :return: The client instance after adding the settings
        """
        if repository:
            self.repository = repository

        if schemas:
            if isinstance(schemas, list):
                schemas = ",".join(schemas)
            self.schemas = schemas

        return self

    def request(
        self,
        method,  # type: str
        path,  # type: str
        headers=None,  # type: Optional[Dict[str, str]]
        data=None,  # type: Optional[Any]
        raw=False,  # type: bool
        ssl_verify=True,  # type: bool
        **kwargs,  # type: Any
    ):
        # type: (...) -> Union[requests.Response, Any]
        """
        Send a request to the Nuxeo server.

        :param method: the HTTP method
        :param path: the path to append to the host
        :param headers: the headers for the HTTP request
        :param data: data to put in the body
        :param raw: if True, don't parse the data to JSON
        :param kwargs: other parameters accepted by
               :func:`requests.request`
        :return: the HTTP response
        """
        if method not in (
            "GET",
            "HEAD",
            "POST",
            "PUT",
            "DELETE",
            "CONNECT",
            "OPTIONS",
            "TRACE",
        ):
            raise BadQuery("method parameter is not a valid HTTP method.")

        # Construct the full URL without double slashes
        url = self.host + path.lstrip("/")
        if "adapter" in kwargs:
            url = f"{url}/@{kwargs.pop('adapter')}"

        kwargs.update(self.client_kwargs)

        # Set the default value to `object` to allow someone
        # to set `timeout` to `None`.
        if kwargs.get("timeout", object) is object:
            kwargs["timeout"] = (TIMEOUT_CONNECT, TIMEOUT_READ)

        headers = headers or {}
        if "Content-Type" not in headers:
            headers["Content-Type"] = kwargs.pop("content_type", "application/json")
        headers.update(
            {"X-NXDocumentProperties": self.schemas, "X-NXRepository": self.repository}
        )
        enrichers = kwargs.pop("enrichers", None)
        if enrichers:
            headers["enrichers-document"] = ", ".join(enrichers)

        headers.update(self.headers)
        self._check_headers_and_params_format(headers, kwargs.get("params") or {})

        if data and not isinstance(data, bytes) and not raw:
            data = json.dumps(data, default=json_helper)

        # Set the default value to `object` to allow someone
        # to set `default` to `None`.
        default = kwargs.pop("default", object)

        # Allow to pass a custom authentication class
        auth = kwargs.pop("auth", None) or self.auth

        _kwargs = {k: v for k, v in kwargs.items() if k != "params"}
        logged_params = kwargs.get("params", {} if raw else data)
        logger.debug(
            (
                f"Calling {method} {url!r} with headers={headers!r},"
                f" params={logged_params!r}, kwargs={_kwargs!r}"
                f" and cookies={self._session.cookies!r}"
            )
        )

        exc = None
        ssl_verify_needed = self.ssl_verify_needed

        if ssl_verify_needed:
            ssl_verify_needed = ssl_verify
        if ssl_verify_needed and "verify" in kwargs:
            ssl_verify_needed = kwargs["verify"]
        kwargs.pop("verify")

        try:
            resp = self._session.request(
                method,
                url,
                headers=headers,
                auth=auth,
                data=data,
                verify=ssl_verify_needed,
                **kwargs,
            )
            resp.raise_for_status()
        except Exception as exc:
            if default is object:
                raise self._handle_error(exc)
            resp = default
        finally:
            # Explicitly break a reference cycle
            exc = None
            del exc

        return resp

    def _check_headers_and_params_format(self, headers, params):
        # type: (Dict[str, Any], Dict[str, Any]) -> None
        """Check headers and params keys for dots or underscores and throw a warning if one is found."""

        msg = "{!r} {} should not contain '_' nor '.'. Replace with '-' to get rid of that warning."

        for key in headers.keys():
            if "_" in key or "." in key:
                warn(msg.format(key, "header"), DeprecationWarning, 2)

        if not isinstance(params, dict):
            return
        for key in params.keys():
            if "_" in key or "." in key:
                warn(msg.format(key, "param"), DeprecationWarning, 2)

    def request_auth_token(
        self,
        device_id,  # type: str
        permission,  # type: str
        app_name=DEFAULT_APP_NAME,  # type: str
        device=None,  # type: Optional[str]
        revoke=False,  # type: bool
        ssl_verify=True,  # type: bool
    ):
        # type: (...) -> str
        """
        Request a token for the user.
        It should only be used if you want to get a Nuxeo token from a Basic Auth.

        :param device_id: device identifier
        :param permission: read/write permissions
        :param app_name: application name
        :param device: optional device description
        :param revoke: revoke the token
        """
        auth = TokenAuth("")
        token = auth.request_token(
            self,
            device_id,
            permission,
            app_name=app_name,
            device=device,
            revoke=revoke,
            auth=self.auth,
            ssl_verify=ssl_verify,
        )

        # Use the (potentially re-newed) token from now on
        if not revoke:
            self.auth = auth
        return token

    def is_reachable(self):
        # type: () -> bool
        """Check if the Nuxeo Platform is reachable."""
        response = self.request("GET", "runningstatus", default=False)
        if isinstance(response, requests.Response):
            return response.ok
        else:
            return bool(response)

    def server_info(self, force=False, ssl_verify=True):
        # type: (bool, bool) -> Dict[str, str]
        """
        Retreive server information.

        :param bool force: Force information renewal.
        """
        if force or self._server_info is None:
            try:
                response = self.request("GET", "json/cmis", ssl_verify=ssl_verify)
                self._server_info = response.json()["default"]
            except Exception:
                logger.warning(
                    "Invalid response data when called server_info()", exc_info=True
                )
        return self._server_info

    @property
    def server_version(self, ssl_verify=True):
        # type: (bool) -> str
        """Return the server version or "unknown"."""
        try:
            return self.server_info(ssl_verify)["productVersion"]
        except Exception:
            return "unknown"

    @staticmethod
    def _handle_error(error):
        # type: (Exception) -> Exception
        """
        Log error and handle raise.

        :param error: The error to handle
        """
        if not isinstance(error, requests.HTTPError):
            return error

        response = error.response
        error_data = {}
        try:
            error_data.update(response.json())
        except ValueError:
            error_data["message"] = response.content
        finally:
            error_data["status"] = response.status_code
            if not error_data.get("message", ""):
                error_data["message"] = response.reason

        status = error_data["status"]
        request_uid = response.headers.get(IDEMPOTENCY_KEY, "")
        if status == 409 and request_uid:
            return OngoingRequestError(request_uid)
        return HTTP_ERROR.get(status, HTTPError).parse(error_data)


class Nuxeo(object):
    """
    Instantiate the client and all the API Endpoints.

    :param auth: the authenticator
    :param host: the host URL
    :param app_name: the name of the application using the client
    :param client: the client class
    :param kwargs: any other argument to forward to every requests calls
    """

    def __init__(
        self,
        auth=None,  # type: Optional[Tuple[str, str]]
        host=DEFAULT_URL,  # type: str
        app_name=DEFAULT_APP_NAME,  # type: str
        version=__version__,  # type: str
        verify=True,  # bool
        client=NuxeoClient,  # type: Type[NuxeoClient]
        **kwargs,  # type: Any
    ):
        # type: (...) -> None
        kwargs["verify"] = verify
        self.client = client(
            auth, host=host, app_name=app_name, version=version, **kwargs
        )
        self.comments = comments.API(self.client)
        self.operations = operations.API(self.client)
        self.directories = directories.API(self.client)
        self.groups = groups.API(self.client)
        self.tasks = tasks.API(self.client)
        self.uploads = uploads.API(self.client)
        self.users = users.API(self.client)
        self.workflows = workflows.API(self.client, self.tasks)
        self.documents = documents.API(
            self.client, self.operations, self.workflows, self.comments
        )

    def __repr__(self):
        # type: () -> str
        return f"{type(self).__name__}<version={__version__!r}, client={self.client!r}>"

    def __str__(self):
        # type: () -> str
        return repr(self)

    def can_use(self, operation):
        # type: (str) -> str
        """Return a boolean to let the caller know if the given *operation* can be used."""
        return operation in self.operations.operations



# coding: utf-8
from typing import TYPE_CHECKING, Dict, Optional

from .endpoint import APIEndpoint
from .models import Group

if TYPE_CHECKING:
    from .client import NuxeoClient


class API(APIEndpoint):
    """Endpoint for groups."""

    __slots__ = ("params",)

    def __init__(self, client, endpoint="group", headers=None):
        # type: (NuxeoClient, str, Optional[Dict[str, str]]) -> None
        self.params = {"fetch-group": ["memberUsers", "memberGroups"]}
        super().__init__(client, endpoint=endpoint, cls=Group, headers=headers)

    def get(self, group_id=None, ssl_verify=True):
        # type: (Optional[str], Optional[bool]) -> Group
        """
        Get the detail of a group.

        :param group_id: the id of the group
        :return: the group
        """
        return super().get(path=group_id, params=self.params, ssl_verify=ssl_verify)

    def post(self, group, ssl_verify=True):
        # type: (Group, bool) -> Group
        """
        Create a group.

        :param group: the group to create
        :return: the created group
        """

        return super().post(resource=group, params=self.params, ssl_verify=ssl_verify)

    create = post  # Alias for clarity

    def put(self, group, ssl_verify=True):
        # type: (Group, bool) -> Group
        """
        Update a group.

        :param group: the group to update
        :return: the updated group
        """
        return super().put(group, ssl_verify=ssl_verify)

    def delete(self, group_id, ssl_verify=True):
        # type: (str, bool) -> None
        """
        Delete a group.

        :param group_id: the id of the group to delete
        """

        super().delete(group_id, ssl_verify=ssl_verify)



# coding: utf-8
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from .endpoint import APIEndpoint
from .models import Document, Task, Workflow
from .tasks import API as TasksAPI

if TYPE_CHECKING:
    from .client import NuxeoClient


class API(APIEndpoint):
    """ Endpoint for workflows. """

    __slots__ = ("tasks_api",)

    def __init__(
        self,
        client,  # type: NuxeoClient
        tasks,  # type: TasksAPI
        endpoint="workflow",  # type: str
        headers=None,  # type: Optional[Dict[str, str]]
    ):
        # type: (...) -> None
        self.tasks_api = tasks
        super().__init__(client, endpoint=endpoint, cls=Workflow, headers=headers)

    def get(self, workflow_id=None):
        # type: (Optional[str]) -> Workflow
        """
        Get the detail of a workflow.

        :param workflow_id: the id of the workflow
        :return: the workflow
        """
        return super().get(path=workflow_id)

    def post(
        self,
        model,  # type: str
        document=None,  # type: Optional[Document]
        options=None,  # type: Optional[Dict[str, Any]]
    ):
        # type: (...) -> Workflow
        """
        Start a workflow.

        :param model: the workflow to start
        :param document: the document to start the workflow on
        :param options: options for the workflow
        :return: the created workflow
        """
        data = {"workflowModelName": model, "entity-type": "workflow"}
        options = options or {}
        if "attachedDocumentIds" in options:
            data["attachedDocumentIds"] = options["attachedDocumentIds"]
        if "variables" in options:
            data["variables"] = options["variables"]

        kwargs = {}
        if document:
            kwargs["endpoint"] = self.client.api_path
            kwargs["path"] = f"id/{document.uid}/@workflow"
        return super().post(data, **kwargs)

    start = post  # Alias for clarity

    def put(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def delete(self, workflow_id):
        # type: (str) -> None
        """
        Delete a workflow.

        :param workflow_id: the id of the workflow to delete
        """
        super().delete(workflow_id)

    def graph(self, workflow):
        # type: (Workflow) -> Dict[str, Any]
        """
        Get the graph of the workflow in JSON format.

        :param workflow: the worklow to get the graph from
        :return: the graph
        """
        request_path = f"{workflow.uid}/graph"
        return super().get(path=request_path)

    def started(self, model):
        # type: (str) -> List[Workflow]
        """
        Get started workflows having the specified model.

        :param model: the workflow model
        :return: the started workflows
        """
        return super().get(params={"workflowModelName": model})

    def tasks(self, workflow):
        # type: (Workflow) -> List[Task]
        """ Get the tasks of a workflow. """
        return self.tasks_api.get(workflow.as_dict())



# coding: utf-8
from requests import Request


class AuthBase(object):
    """Base class that all auth implementations derive from.
    Copied from requests.auth to add __slots__ capability.
    """

    __slots__ = ()

    def __call__(self, r):
        # type: (Request) -> Request
        raise NotImplementedError("Auth hooks must be callable.")
